/// TKS_Fork.h 
/// 
/// (c) 2005 Bastian Spiegel <bs@tkscript.de> 
///     - distributed under terms of the GNU general public license (GPL). 
/// 
 
#ifndef __TKS_FORK_H__
#define __TKS_FORK_H__


enum __tks_fork_ipc_offsets {
   TKS_FORK_IPC_ALIVE    = 0,   // sU32        !=0 if the parent<->child link has been established.
   TKS_FORK_IPC_STATUS   = 4,   // sS32        see TKS_FORK_STATUS_xxx
   TKS_FORK_IPC_ATOM     = 8,   // sS32        the current atomic type of the transfer buffer, see YAC_TYPE_xxx.
   TKS_FORK_IPC_FUNC     = 12,  // sChar[244]  ASCIIZ terminated name of called function (244 chars)
   TKS_FORK_IPC_TRANSFER = 256, // sU8*        serialized or raw data argument (status=call/exec) or call return value (status=return) 
   
   TKS_FORK_DEFAULT_TRANSFER_SIZE   = 16140
};


enum __tks_fork_status_codes {
   TKS_FORK_STATUS_IDLE    = 0, // parent process has retrieved the last call result and is ready to post next request
   TKS_FORK_STATUS_CALL    = 1, // parent process signalizes 
   TKS_FORK_STATUS_EXEC    = 2, // child process signalizes execution of the current call request. the parent process may reset the status to IDLE to signalize the child process to cancel the current call.
   TKS_FORK_STATUS_RETURN  = 3, // child process signalizes successful execution of call request
   TKS_FORK_STATUS_FAIL    = 4, // child process signalizes error while executing call request
   TKS_FORK_STATUS_DIE     = 5, // child process signalizes that it had to shutdown (segv, ill, internal (caught) error etc..) and now wants to be restarted

   TKS_FORK_STATUS_PAD
};


class TKS_Fork : public YAC_Object {
public:
   YAC_String chapter_name; // the .tkx/tkp chapter which describes the worker script-"thread"

   void *process_handle; // OS-specific; represents the worker-process

   TKS_SharedBuffer ipc_buffer; // used for inter process communication, see __tks_fork_ipc_offsets.

   YAC_Buffer transfer_buffer; // the IPC_TRANSFER raw data; a view into the ipc_buffer.

public:
   TKS_Fork(void);
   ~TKS_Fork();
   
   void YAC_VCALL yacGetConstantStringList(YAC_String *); // FORK_STATUS_xxx
   
   void free(void); // terminate the worker-process if it is still running (and wait until it has exited), free shared memory
   static sChar *getStatusStringByCode(sSI _errorCode);

   void        abort           (void); // signalize child to abort the current call.
   sSI         call            (YAC_String *_function, YAC_Object *_arg); // wait for last call to finish; call worker-function; return true if call has been encoded to ipc_buffer. _arg==null => leave transfer buffer unchanged.
   YAC_Object *getBuffer       (void); // returns a reference to the transfer buffer
   sSI         getBufferOffset (void); // query the current i/o offset
   sSI         getBufferSize   (void); // query the size of the transfer buffer
   YAC_String *getName         (void); // return chapter name
   sSI         getPriority     (void); // query priority of child process
   void        getResult       (YAC_Value *_r); // try to deserialize result value from the transfer buffer, wait until status!=exec. returns void if call has failed or value could not be deserialized.
   sSI         getStatusCode   (void); // see TKS_FORK_STATUS_xxx
   void        getStatusString (YAC_String *_r); // "idle", "call", "return", "fail"
   sSI         haveResult      (void); // query whether current call already returned a result.
   void        isAlive         (void); // query whether the parent<->child shared memory link has been established.
   sSI         isBusy          (void); // query whether status is !=exec.
   void        setBufferOffset (sSI);  // set the current i/o offset. this can be used to offset serialized data by a (fixed) raw-data buffer header.
   void        setBufferSize   (sSI);  // set the size of the transfer buffer.
   void        setPriority     (sSI);  // set the priority of the child process -10 (low)..10 (high)
   sSI         start           (void); // actually start the new tks child process. return true if the new process has been created. do not wait for process to initialize.
   sSI         startAndWait    (sSI _timeoutMilliSeconds); // start new tks child process and wait until the shared memory link has been established. return true if the new process has been created and the shm-link has been established.
   sSI         wait            (sSI _timeoutMilliSeconds); // wait for child process to become alive resp. change status code to !=exec. return false if call timed out.
};


#endif // __TKS_FORK_H__
